# -*- coding: utf-8 -*-
"""Constant dimensions to allow for human access within a plant."""

from aguaclara.core.units import u

WALKWAY_W_DEFAULT = 1 * u.m
DRAIN_CHAN_WALKWAY_W = 1.2 * u.m
BASEMENT_STAIRS_W = 0.9 * u.m
ENTRANCE_STAIRS_W = 1.2 * u.m

HUMAN_W_MIN = 45 * u.cm
DRAIN_CHAN_H_MIN = 1.5 * u.m
PLANT_H_MIN = 0.1 * u.m

DOOR_W = 1 * u.m
